@@ -20,7 +20,6 @@ import io
 import logging
 import os
 import re
-import sys
 import textwrap
 import enum
 import urllib.parse
@@ -30,7 +29,6 @@ from typing import TYPE_CHECKING, Dict, List, Iterator, Tuple, Optional, Set, Te
 
 from zc.buildout.buildout import _buildout_default_options
 from zc.buildout.configparser import (
-    Error,
     MissingSectionHeaderError,
     ParsingError,
     leading_blank_lines,
@@ -42,7 +40,8 @@ from . import recipes
 from . import jinja
 
 from pygls.server import LanguageServer
-from pygls.types import Position, Range, Location
+from pygls.workspace import Document
+from pygls.lsp.types import Position, Range, Location
 
 logger = logging.getLogger(__name__)
 requests_session = requests.Session()
@@ -106,21 +105,69 @@ class MissingExtendedSection(ResolveError):
 
 
 class BuildoutOptionDefinition:
+  """Option definition
+
+  Keep track of the current value as `value` and of the
+  `locations` where this value was defined. `values` is
+  a list of intermediate values for each of the locations.
+
+  `default_value` are default values that are not defined
+  in profiles, but are implicit, such as buildout default
+  values or sections added by slapos instance.
+  """
   def __init__(
       self,
-      locations: List[Location],
       value: str,
-      implicit_option: bool = False,
+      location: Location,
+      default_value: bool = False,
   ):
-    self.locations = locations
-    self.value = value
-    self.implicit_option = implicit_option
+    self.locations: Tuple[Location, ...] = (location, )
+    self.values: Tuple[str, ...] = (value, )
+    self.default_values: Tuple[bool, ...] = (default_value, )
+
+  @property
+  def value(self) -> str:
+    return self.values[-1]
+
+  @property
+  def location(self) -> Location:
+    return self.locations[-1]
+
+  @property
+  def default_value(self) -> bool:
+    return self.default_values[-1]
 
   def __repr__(self) -> str:
     locations = ' '.join(
         ['{} {}'.format(l.uri, l.range) for l in self.locations])
     return '{} ({})'.format(self.value, locations)
 
+  def overrideValue(self, value: str, location: Location) -> None:
+    """Add a value to the list of values."""
+    self.values = self.values + (value, )
+    self.locations = self.locations + (location, )
+    self.default_values = self.default_values + (False, )
+
+  def updateValue(
+      self,
+      value: str,
+      location: Optional[Location] = None,
+  ) -> None:
+    """Replace the current value, used internally to clean up extra whitespaces."""
+    self.values = self.values[:-1] + (value, )
+    self.default_values = self.default_values[:-1] + (False, )
+    if location is not None:
+      self.locations = self.locations[:-1] + (location, )
+
+  @staticmethod
+  def clone(
+      option_def: 'BuildoutOptionDefinition') -> 'BuildoutOptionDefinition':
+    new_option_def = BuildoutOptionDefinition(option_def.value,
+                                              option_def.locations[0])
+    new_option_def.locations = option_def.locations
+    new_option_def.values = option_def.values
+    return new_option_def
+
 
 class _BuildoutSection(Dict[str, BuildoutOptionDefinition]):
   """Section of a buildout.
@@ -400,23 +447,23 @@ class BuildoutTemplate:
             is_same_section_reference=match.group('section') == '',
         )
         symbol.section_range = Range(
-            Position(
-                lineno,
-                match.start() + 2,  # the ${ was captured
+            start=Position(
+                line=lineno,
+                character=match.start() + 2,  # the ${ was captured
             ),
-            Position(
-                lineno,
-                match.end() - len(match.group('option')) - 1,
+            end=Position(
+                line=lineno,
+                character=match.end() - len(match.group('option')) - 1,
             ),
         )
         symbol.option_range = Range(
-            Position(
-                lineno,
-                match.end() - len(match.group('option')),
+            start=Position(
+                line=lineno,
+                character=match.end() - len(match.group('option')),
             ),
-            Position(
-                lineno,
-                match.end(),
+            end=Position(
+                line=lineno,
+                character=match.end(),
             ),
         )
         yield symbol
@@ -436,8 +483,9 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
                                         Location] = collections.OrderedDict()
     """The locations for each section, keyed by section names.
     """
-    self.extends_jinja = False
-    """Flag true if this resolved buildout had extends defined by jinja expressions.
+    self.has_dynamic_extends = False
+    """Flag true if this resolved buildout had extends defined dynamically.
+    This only happens with SlapOS instance buildout which are templates of profiles.
     """
 
   async def getTemplate(
@@ -722,13 +770,13 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
               yield (
                   option_value_text,
                   Range(
-                      Position(
-                          start_line + line_offset,
-                          start_character,
+                      start=Position(
+                          line=start_line + line_offset,
+                          character=start_character,
                       ),
-                      Position(
-                          start_line + line_offset,
-                          start_character + len(option_value_text),
+                      end=Position(
+                          line=start_line + line_offset,
+                          character=start_character + len(option_value_text),
                       ),
                   ),
               )
@@ -736,13 +784,13 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
               for match in re.finditer(r'([^\s]+)', option_value_text):
                 yield (match.group(),
                        Range(
-                           Position(
-                               start_line + line_offset,
-                               start_character + match.start(),
+                           start=Position(
+                               line=start_line + line_offset,
+                               character=start_character + match.start(),
                            ),
-                           Position(
-                               start_line + line_offset,
-                               start_character + match.start() +
+                           end=Position(
+                               line=start_line + line_offset,
+                               character=start_character + match.start() +
                                len(match.group()),
                            ),
                        ))
@@ -881,23 +929,32 @@ async def _parse(
       value = v.value
     sections['buildout'][k] = BuildoutOptionDefinition(
         value=value,
-        locations=[Location(uri=uri, range=Range(Position(0), Position(0)))],
-        implicit_option=True,
+        location=Location(uri=uri,
+                          range=Range(start=Position(line=0, character=0),
+                                      end=Position(line=0, character=0))),
+        default_value=True,
     )
   sections['buildout']['directory'] = BuildoutOptionDefinition(
       value='.',
-      locations=[Location(uri=uri, range=Range(Position(0), Position(0)))],
-      implicit_option=True,
+      location=Location(uri=uri,
+                        range=Range(start=Position(line=0, character=0),
+                                    end=Position(line=0, character=0))),
+      default_value=True,
   )
   sections.section_header_locations['buildout'] = Location(
       uri="",
-      range=Range(Position(), Position()),
+      range=Range(
+          start=Position(line=0, character=0),
+          end=Position(line=0, character=0),
+      ),
   )
   if slapos_instance_profile_filename_re.match(uri):
     # Add slapos instance generated sections.
     sections.section_header_locations.setdefault(
-        'slap-connection', Location(uri='',
-                                    range=Range(Position(), Position())))
+        'slap-connection',
+        Location(uri='',
+                 range=Range(start=Position(line=0, character=0),
+                             end=Position(line=0, character=0))))
     slap_connection = BuildoutSection()
     for k in (
         'computer-id',
@@ -908,14 +965,18 @@ async def _parse(
         'software-release-url',
     ):
       slap_connection[k] = BuildoutOptionDefinition(
-          locations=[],
           value='',
-          implicit_option=True,
+          location=Location(uri=uri,
+                            range=Range(start=Position(line=0, character=0),
+                                        end=Position(line=0, character=0))),
+          default_value=True,
       )
     sections.setdefault('slap-connection', slap_connection)
     sections.section_header_locations.setdefault(
         'slap-network-information',
-        Location(uri='', range=Range(Position(), Position())))
+        Location(uri='',
+                 range=Range(start=Position(line=0, character=0),
+                             end=Position(line=0, character=0))))
     slap_network_information = BuildoutSection()
     for k in (
         'local-ipv4',
@@ -928,9 +989,11 @@ async def _parse(
         'global-ipv4-network',
     ):
       slap_network_information[k] = BuildoutOptionDefinition(
-          locations=[],
           value='',
-          implicit_option=True,
+          location=Location(uri=uri,
+                            range=Range(start=Position(line=0, character=0),
+                                        end=Position(line=0, character=0))),
+          default_value=True,
       )
     sections.setdefault('slap-network-information', slap_network_information)
 
@@ -966,9 +1029,17 @@ async def _parse(
       assert cursect is not None
       assert optname is not None
       option_def = cursect[optname]
-      option_def.locations[-1].range.end.line = lineno
-      option_def.locations[-1].range.end.character = len(_line) - 1
-      option_def.value = ("%s\n%s" % (option_def.value, line))
+      # update current option in case of multi line option
+      option_def.updateValue(
+          value=("%s\n%s" % (option_def.value, line)),
+          location=Location(
+              uri=option_def.location.uri,
+              range=Range(
+                  start=option_def.location.range.start,
+                  end=Position(line=lineno, character=len(_line) - 1),
+              ),
+          ),
+      )
       cursect[optname] = option_def
 
     else:
@@ -978,8 +1049,8 @@ async def _parse(
         sections.section_header_locations[sectname] = Location(
             uri=uri,
             range=Range(
-                Position(lineno, 0),
-                Position(lineno + 1, 0),
+                start=Position(line=lineno, character=0),
+                end=Position(line=lineno + 1, character=0),
             ))
         if sectname in sections:
           cursect = sections[sectname]
@@ -987,24 +1058,26 @@ async def _parse(
           sections[sectname] = cursect = BuildoutSection()
           # initialize buildout default options
           cursect['_buildout_section_name_'] = BuildoutOptionDefinition(
-              locations=[
-                  Location(uri=uri,
-                           range=Range(Position(0, 0), Position(0, 0)))
-              ],
+              location=Location(uri=uri,
+                                range=Range(start=Position(line=0,
+                                                           character=0),
+                                            end=Position(line=0,
+                                                         character=0))),
               value=sectname,
-              implicit_option=True,
+              default_value=True,
           )
           # _profile_base_location_ is a slapos.buildout extension
           base_location = '.'
           if '/' in uri:
             base_location = uri[:uri.rfind('/')] + '/'
           cursect['_profile_base_location_'] = BuildoutOptionDefinition(
-              locations=[
-                  Location(uri=uri,
-                           range=Range(Position(0, 0), Position(0, 0)))
-              ],
+              location=Location(uri=uri,
+                                range=Range(start=Position(line=0,
+                                                           character=0),
+                                            end=Position(line=0,
+                                                         character=0))),
               value=base_location,
-              implicit_option=True,
+              default_value=True,
           )
 
         # So sections can't start with a continuation line
@@ -1026,18 +1099,25 @@ async def _parse(
           assert optname
           optname = optname.rstrip()
           optval = optval.strip()
-          option_def = cursect.get(
-              optname, BuildoutOptionDefinition(value=optval, locations=[]))
-          option_def.value = optval
-
-          option_def.locations.append(
-              Location(uri=uri,
-                       range=Range(
-                           start=Position(line=lineno,
-                                          character=len(mo.groups()[0]) + 1),
-                           end=Position(line=lineno, character=len(line) - 1),
-                       )))
-          option_def.implicit_option = False
+          optlocation = Location(
+              uri=uri,
+              range=Range(
+                  start=Position(
+                      line=lineno,
+                      character=len(mo.groups()[0]) + 1,
+                  ),
+                  end=Position(
+                      line=lineno,
+                      character=len(line) - 1,
+                  ),
+              ),
+          )
+          if optname in cursect:
+            option_def = cursect[optname]
+            option_def.overrideValue(optval, optlocation)
+          else:
+            option_def = BuildoutOptionDefinition(value=optval,
+                                                  location=optlocation)
           cursect[optname] = option_def
           blockmode = not optval
         elif not (optname or line.strip()):
@@ -1061,12 +1141,69 @@ async def _parse(
     for name in section:
       value = section[name].value
       if value[:1].isspace():
-        section[name].value = leading_blank_lines.sub(
-            '', textwrap.dedent(value.rstrip()))
+        section[name].updateValue(
+            leading_blank_lines.sub('', textwrap.dedent(value.rstrip())))
 
   return sections
 
 
+async def getProfileForTemplate(
+    ls: LanguageServer,
+    document: Document,
+) -> Optional[URI]:
+  """Find the profile for template.
+
+  For example when there's a buildout.cfg containing:
+  
+    [template]
+    recipe = collective.recipe.template
+    input = template.in
+    output = template
+
+  when called with `uri` template.in, this function would return `buildout.cfg`.
+
+  """
+  uri = document.uri
+
+  def getCandidateBuildoutProfiles() -> Iterator[pathlib.Path]:
+    path = pathlib.Path(document.path).parent
+    for _ in range(3):  # look for buildouts up to 3 levels
+      # we sort just to have stable behavior
+      for profile in sorted(path.glob('*.cfg')):
+        yield profile
+      path = path.parent
+
+  if slapos_instance_profile_filename_re.match(
+      uri) or not uri.endswith('.cfg'):
+    for buildout_path in getCandidateBuildoutProfiles():
+      resolved_path = str(buildout_path.resolve())
+      # For paths in workspace, we don't use buildout_path.resolve().as_uri(),
+      # because we have fake uri -> path mapping in tests
+      if resolved_path.startswith(ls.workspace.root_path):
+        buildout_uri = resolved_path.replace(
+            ls.workspace.root_path,
+            ls.workspace.root_uri,
+            1,
+        )
+      else:
+        # but we still need to support the case where the path is outside the workspace
+        buildout_uri = buildout_path.resolve().as_uri()
+      logger.debug("Trying to find templates's buildout with %s -> %s",
+                   buildout_path, buildout_uri)
+      buildout = await _open(
+          ls,
+          '',
+          buildout_uri,
+          [],
+          allow_errors=True,
+      )
+      assert isinstance(buildout, BuildoutProfile)
+      template = await buildout.getTemplate(ls, uri)
+      if template is not None:
+        return buildout_uri
+  return None
+
+
 async def open(
     ls: LanguageServer,
     uri: URI,
@@ -1087,48 +1224,20 @@ async def open(
   document = ls.workspace.get_document(uri)
   logger.debug("open %s", uri)
   if not force_open_as_buildout_profile:
-
-    def getCandidateBuildoutProfiles() -> Iterator[pathlib.Path]:
-      path = pathlib.Path(document.path).parent
-      for _ in range(3):  # look for buildouts up to 3 levels
-        # we sort just to have stable behavior
-        for profile in sorted(path.glob('*.cfg')):
-          yield profile
-        path = path.parent
-
     # First, try to read as a template, because buildout profiles can be templates.
-    if slapos_instance_profile_filename_re.match(
-        uri) or not uri.endswith('.cfg'):
-      for buildout_path in getCandidateBuildoutProfiles():
-        resolved_path = str(buildout_path.resolve())
-        # For paths in workspace, we don't use buildout_path.resolve().as_uri(),
-        # because we have fake uri -> path mapping in tests
-        if resolved_path.startswith(ls.workspace.root_path):
-          buildout_uri = resolved_path.replace(
-              ls.workspace.root_path,
-              ls.workspace.root_uri,
-              1,
-          )
-        else:
-          # but we still need to support the case where the path is outside the workspace
-          buildout_uri = buildout_path.resolve().as_uri()
-        logger.debug("Trying to find templates's buildout with %s -> %s",
-                     buildout_path, buildout_uri)
-        buildout = await _open(
-            ls,
-            '',
-            buildout_uri,
-            [],
-            allow_errors=allow_errors,
-        )
-        assert isinstance(buildout, BuildoutProfile)
-        template = await buildout.getTemplate(ls, uri)
-        if template is not None:
-          return template
+    buildout_uri = await getProfileForTemplate(ls, document)
+    if buildout_uri is not None:
+      buildout = await _open(
+          ls,
+          '',
+          buildout_uri,
+          [],
+          allow_errors=allow_errors,
+      )
+      return await buildout.getTemplate(ls, uri)
 
   if BuildoutProfile.looksLikeBuildoutProfile(
       uri) or force_open_as_buildout_profile:
-    fp = io.StringIO(document.source)
     return await _open(ls, '', uri, [], allow_errors=allow_errors)
 
   return None
@@ -1162,22 +1271,23 @@ async def _open(
       return ResolvedBuildout(uri, '')
     raise RecursiveIncludeError("Recursive file include", seen, uri)
 
-  root_config_file = not seen
   seen.append(uri)
 
-  non_extended = result = await parse(ls, uri, allow_errors=allow_errors)
+  result = await parse(ls, uri, allow_errors=allow_errors)
 
   extends_option = result['buildout'].pop(
       'extends', None) if 'buildout' in result else None
-  extends_jinja = False
+  has_dynamic_extends = False
   if extends_option:
     extends = extends_option.value.split()
-    extends_jinja = (jinja.JinjaParser.jinja_value in extends)
+    has_dynamic_extends = (jinja.JinjaParser.jinja_value in extends) or any(
+        option_reference_re.match(extended_profile)
+        for extended_profile in extends)
     if extends:
       # extends, as absolute URI that we can use as cache key
       absolute_extends = tuple(urllib.parse.urljoin(base, x) for x in extends)
       eresult = await _open(ls, base, extends.pop(0), seen, allow_errors)
-      extends_jinja = extends_jinja or eresult.extends_jinja
+      has_dynamic_extends = has_dynamic_extends or eresult.has_dynamic_extends
       for fname in extends:
         _update(eresult, await _open(ls, base, fname, seen, allow_errors))
       for absolute_extend in absolute_extends:
@@ -1200,7 +1310,7 @@ async def _open(
         # this happens with non top-level buildout
         pass
 
-  result.extends_jinja = extends_jinja
+  result.has_dynamic_extends = has_dynamic_extends
   resolved = cast(ResolvedBuildout, result)
   _resolved_buildout_cache[uri] = resolved
   return copy.deepcopy(resolved)
@@ -1212,29 +1322,39 @@ def _update_section(
 ) -> BuildoutSection:
   """Update s1 with values from s2.
   """
-  s2 = copy.deepcopy(s2)
+  s2 = copy.copy(s2)
   for k, v in s2.items():
+    if k == '_profile_base_location_':
+      continue
     if k.endswith('-'):
       k = k.rstrip(' -')
       # Find v1 in s2 first; it may have been set by a += operation first
       option_def = s2.get(k, s1.get(k, v))
-      # same logic as as SectionKey.removeFromValue
-      option_def.value = '\n'.join(new_v
-                                   for new_v in option_def.value.split('\n')
-                                   if new_v not in v.value.split('\n'))
-      option_def.locations.extend(v.locations)
-      s1[k] = option_def
+      new_option_def = BuildoutOptionDefinition.clone(option_def)
+      new_option_def.overrideValue(
+          # same logic as as SectionKey.removeFromValue
+          value='\n'.join(new_v for new_v in option_def.value.split('\n')
+                          if new_v not in v.value.split('\n')),
+          location=v.locations[-1])
+      s1[k] = new_option_def
     elif k.endswith('+'):
       k = k.rstrip(' +')
       # Find v1 in s2 first; it may have been defined locally too.
       option_def = s2.get(k, s1.get(k, v))
-      # same logic as as SectionKey.addToValue
-      option_def.value = '\n'.join(
-          option_def.value.split('\n') + v.value.split('\n'))
-      option_def.locations.extend(v.locations)
-      s1[k] = option_def
+      option_values = [] if option_def.default_value else option_def.value.split(
+          '\n')
+      new_option_def = BuildoutOptionDefinition.clone(option_def)
+      new_option_def.overrideValue(
+          # same logic as as SectionKey.addToValue
+          value='\n'.join(option_values + v.value.split('\n')),
+          location=v.location)
+      s1[k] = new_option_def
     else:
-      s1[k] = v
+      if k in s1 and (v.location != s1[k].location):
+        if not v.default_value:
+          s1[k].overrideValue(v.value, v.location)
+      else:
+        s1[k] = v
   return s1
 
 
@@ -1276,7 +1396,6 @@ def _do_extend_raw(
     to_do = section.get('<', None)
     if to_do is None:
       return section
-    __doing__ = 'Loading input sections for %r', name
 
     result = BuildoutSection()
     for iname in to_do.value.split('\n'):
@@ -1286,9 +1405,11 @@ def _do_extend_raw(
       raw = buildout.get(iname)
       if raw is None:
         raise MissingExtendedSection("No section named %r" % iname)
-      result.update(_do_extend_raw(iname, raw, buildout, doing))
-
-    _update_section(result, copy.deepcopy(section))
+      result.update({
+          k: copy.copy(v)
+          for (k, v) in _do_extend_raw(iname, raw, buildout, doing).items()
+      })
+    _update_section(result, section)
     result.pop('<', None)
     return result
   finally:
