@@ -4,7 +4,7 @@
 import re
 import enum
 
-from typing import List, Dict
+from typing import List
 
 
 # https://jinja.palletsprojects.com/en/2.11.x/templates/#list-of-control-structures
@@ -29,6 +29,10 @@ end_block_statement = {
 
 statement_re = re.compile(r'.*\{%[\-\+\s]*(?P<statement>[\w]+).*%\}')
 expression_re = re.compile(r'\{\{.*?\}\}')
+multiline_expression_start_re = re.compile(r'^\s*\{\{')
+multiline_expression_end_re = re.compile(r'\}\}')
+multiline_statement_start_re = re.compile(r'^\s*\{%')
+multiline_statement_end_re = re.compile(r'%\}')
 
 
 class JinjaParser:
@@ -45,6 +49,8 @@ class JinjaParser:
     self._stack: List[JinjaStatement] = []
     self._current_line_was_in_jinja = False
     self._in_comment = False
+    self._in_multiline_expression = False
+    self._in_multiline_statement = False
     self.line = ""
 
   def feed(self, line: str) -> None:
@@ -52,8 +58,13 @@ class JinjaParser:
     """
     self._current_line_was_in_jinja = False
     self.has_expression = bool(expression_re.search(line))
-    if expression_re.search(line):
-      line = expression_re.sub(self.jinja_value, line)
+    expression_re_match = expression_re.search(line)
+    if expression_re_match:
+      if expression_re_match.start() == 0 \
+         and expression_re_match.end() == len(line.strip()):
+        line = f'{self.jinja_value} = {self.jinja_value}'
+      else:
+        line = expression_re.sub(self.jinja_value, line)
     self.line = line
     self.is_error = False
 
@@ -73,6 +84,17 @@ class JinjaParser:
           popped = self._stack.pop()
           self.is_error = end_block_statement[statement] != popped
 
+    if multiline_expression_start_re.match(
+        line) or self._in_multiline_expression:
+      self._current_line_was_in_jinja = True
+      self._in_multiline_expression = multiline_expression_end_re.search(
+          line) is None
+    if multiline_statement_start_re.match(
+        line) or self._in_multiline_statement:
+      self._current_line_was_in_jinja = True
+      self._in_multiline_statement = multiline_statement_end_re.search(
+          line) is None
+
   @property
   def is_in_jinja(self) -> bool:
     return (bool(self._stack) or self._current_line_was_in_jinja)
