@@ -1,20 +1,13 @@
 import asyncio
-import contextlib
-import io
 import itertools
 import logging
 import os
 import pathlib
 import re
-import tempfile
 import urllib.parse
-from typing import Any, Iterable, List, Optional, Sequence, Set, Tuple, Union
+from typing import Iterable, List, Optional, Tuple, Union
 
-from zc.buildout import configparser
-from zc.buildout.buildout import Buildout
-from zc.buildout.configparser import MissingSectionHeaderError, ParsingError
-
-from pygls.features import (
+from pygls.lsp.methods import (
     COMPLETION,
     DEFINITION,
     DOCUMENT_LINK,
@@ -23,21 +16,17 @@ from pygls.features import (
     REFERENCES,
     TEXT_DOCUMENT_DID_CHANGE,
     TEXT_DOCUMENT_DID_OPEN,
-    TEXT_DOCUMENT_DID_SAVE,
     WORKSPACE_DID_CHANGE_WATCHED_FILES,
 )
 from pygls.server import LanguageServer
-from pygls.types import (
+from pygls.lsp.types import (
     CompletionItem,
     CompletionItemKind,
-    CompletionList,
+    CompletionOptions,
     CompletionParams,
-    Diagnostic,
-    DiagnosticSeverity,
     DidChangeTextDocumentParams,
-    DidChangeWatchedFiles,
+    DidChangeWatchedFilesParams,
     DidOpenTextDocumentParams,
-    DidSaveTextDocumentParams,
     DocumentLink,
     DocumentLinkParams,
     DocumentSymbol,
@@ -46,24 +35,21 @@ from pygls.types import (
     Location,
     MarkupContent,
     MarkupKind,
-    MessageType,
     Position,
     Range,
     SymbolKind,
     TextDocumentPositionParams,
     TextEdit,
-    WorkspaceEdit,
 )
 from pygls.workspace import Document
 
-from . import buildout, jinja, recipes
+from . import buildout, diagnostic, recipes
 
 server = LanguageServer()
 
 reference_start = '${'
 reference_re = re.compile(
     r'\${(?P<section>[-a-zA-Z0-9 ._]*):(?P<option>[-a-zA-Z0-9 ._]+)}')
-
 logger = logging.getLogger(__name__)
 
 DEBOUNCE_DELAY = 0.3
@@ -84,146 +70,10 @@ async def parseAndSendDiagnostics(
     uri: str,
 ) -> None:
   await asyncio.sleep(DEBOUNCE_DELAY)
-  diagnostics: List[Diagnostic] = []
-  parsed = None
-
-  looks_like_profile = buildout.BuildoutProfile.looksLikeBuildoutProfile(uri)
-
-  if looks_like_profile:
-    # parse errors
-    try:
-      parsed = await buildout.parse(
-          ls=ls,
-          uri=uri,
-          allow_errors=False,
-      )
-    except ParsingError as e:
-      if e.filename != uri:
-        logger.debug("skipping error in external file %s", e.filename)
-      elif isinstance(e, MissingSectionHeaderError):
-        if looks_like_profile:
-          diagnostics.append(
-              Diagnostic(
-                  message=e.message,
-                  range=Range(
-                      start=Position(e.lineno, 0),
-                      end=Position(e.lineno + 1, 0),
-                  ),
-                  source='buildout',
-              ))
-      else:
-        if looks_like_profile:
-          for (lineno, _), msg in zip(e.errors, e.message.splitlines()[1:]):
-            msg = msg.split(':', 1)[1].strip()
-            diagnostics.append(
-                Diagnostic(
-                    message=f'ParseError: {msg}',
-                    range=Range(
-                        start=Position(lineno, 0),
-                        end=Position(lineno + 1, 0),
-                    ),
-                    source='buildout',
-                ))
-
-  resolved_buildout = await buildout.open(
-      ls=ls,
-      uri=uri,
-  )
-  assert resolved_buildout is not None
-
-  installed_parts: Set[str] = set([])
-  if isinstance(resolved_buildout, buildout.BuildoutProfile):
-    if 'parts' in resolved_buildout['buildout']:
-      installed_parts = set(
-          (v[0]
-           for v in resolved_buildout.getOptionValues('buildout', 'parts')))
-
-  async for symbol in resolved_buildout.getAllOptionReferenceSymbols():
-    if symbol.referenced_section is None:
-      diagnostics.append(
-          Diagnostic(
-              message=
-              f'Section `{symbol.referenced_section_name}` does not exist.',
-              range=symbol.section_range,
-              source='buildout',
-          ))
-    elif symbol.referenced_option is None:
-      # if we have a recipe, either it's a known recipe where we know
-      # all options that this recipe can generate, or it's an unknown
-      # recipe and in this case we assume it's OK.
-      if (symbol.referenced_section_recipe_name is not None
-          and symbol.referenced_section_recipe is None) or (
-              symbol.referenced_section_recipe is not None
-              and symbol.referenced_option_name
-              in symbol.referenced_section_recipe.generated_options):
-        continue
-      # if a section is a macro, it's OK to self reference ${:missing}
-      if symbol.is_same_section_reference and symbol.current_section_name not in installed_parts:
-        continue
-      diagnostics.append(
-          Diagnostic(
-              message=
-              f'Option `{symbol.referenced_option_name}` does not exist in `{symbol.referenced_section_name}`.',
-              range=symbol.option_range,
-              source='buildout',
-              severity=DiagnosticSeverity.Warning,
-          ))
-
-  if isinstance(resolved_buildout, buildout.BuildoutProfile):
-    for section_name, section in resolved_buildout.items():
-      if section_name in installed_parts and resolved_buildout.section_header_locations[
-          section_name].uri == uri:
-        # check for required options
-        recipe = section.getRecipe()
-        if recipe:
-          missing_required_options = recipe.required_options.difference(
-              section.keys())
-          if missing_required_options:
-            missing_required_options_text = ', '.join(
-                ['`{}`'.format(o) for o in missing_required_options])
-            diagnostics.append(
-                Diagnostic(
-                    message=
-                    f'Missing required options for `{recipe.name}`: {missing_required_options_text}',
-                    range=resolved_buildout.
-                    section_header_locations[section_name].range,
-                    source='buildout',
-                    severity=DiagnosticSeverity.Error,
-                ), )
-
-    if 'parts' in resolved_buildout['buildout']:
-      jinja_parser = jinja.JinjaParser()
-      for part_name, part_range in resolved_buildout.getOptionValues(
-          'buildout', 'parts'):
-        if part_name:
-          if part_name.startswith('${'):
-            continue  # assume substitutions are OK
-          jinja_parser.feed(part_name)
-          if jinja_parser.is_in_jinja:
-            continue  # ignore anything in jinja context
-
-          if part_name not in resolved_buildout:
-            if not resolved_buildout.extends_jinja:
-              diagnostics.append(
-                  Diagnostic(
-                      message=f'Section `{part_name}` does not exist.',
-                      range=part_range,
-                      source='buildout',
-                      severity=DiagnosticSeverity.Error,
-                  ), )
-          elif 'recipe' not in resolved_buildout[part_name]:
-            diagnostics.append(
-                Diagnostic(
-                    message=f'Section `{part_name}` has no recipe.',
-                    range=part_range,
-                    source='buildout',
-                    severity=DiagnosticSeverity.Error,
-                ), )
-
-  ls.publish_diagnostics(
-      uri,
-      diagnostics,
-  )
+  diagnostics = []
+  async for diag in diagnostic.getDiagnostics(ls, uri):
+    diagnostics.append(diag)
+  ls.publish_diagnostics(uri, diagnostics)
 
 
 @server.feature(TEXT_DOCUMENT_DID_OPEN)
@@ -231,7 +81,7 @@ async def did_open(
     ls: LanguageServer,
     params: DidOpenTextDocumentParams,
 ) -> None:
-  await parseAndSendDiagnostics(ls, params.textDocument.uri)
+  await parseAndSendDiagnostics(ls, params.text_document.uri)
 
 
 @server.feature(TEXT_DOCUMENT_DID_CHANGE)
@@ -239,14 +89,14 @@ async def did_change(
     ls: LanguageServer,
     params: DidChangeTextDocumentParams,
 ) -> None:
-  buildout.clearCache(params.textDocument.uri)
-  await parseAndSendDiagnostics(ls, params.textDocument.uri)
+  buildout.clearCache(params.text_document.uri)
+  await parseAndSendDiagnostics(ls, params.text_document.uri)
 
 
 @server.feature(WORKSPACE_DID_CHANGE_WATCHED_FILES)
 async def did_change_watched_file(
     ls: LanguageServer,
-    params: DidChangeWatchedFiles,
+    params: DidChangeWatchedFilesParams,
 ) -> None:
   for change in params.changes:
     buildout.clearCache(change.uri)
@@ -262,23 +112,26 @@ async def lsp_symbols(
 
   parsed = await buildout.parse(
       ls=ls,
-      uri=params.textDocument.uri,
+      uri=params.text_document.uri,
       allow_errors=True,
   )
 
   for section_name, section_value in parsed.items():
     section_header_location = parsed.section_header_locations[section_name]
     # don't include implicit sections such as [buildout] unless defined in this profile.
-    if section_header_location.uri != params.textDocument.uri:
+    if section_header_location.uri != params.text_document.uri:
       continue
     children: List[DocumentSymbol] = []
     for option_name, option_value in section_value.items():
-      if option_value.implicit_option:
+      if option_value.default_value:
         continue
       option_range = Range(
-          start=Position(
-              min(r.range.start.line for r in option_value.locations)),
-          end=Position(max(r.range.end.line for r in option_value.locations)))
+          start=Position(line=min(r.range.start.line
+                                  for r in option_value.locations),
+                         character=0),
+          end=Position(line=max(r.range.end.line
+                                for r in option_value.locations),
+                       character=0))
       detail = getOptionValue(option_value)
       if len(detail.splitlines()) > 1:
         #  vscode does not like too long multi-lines detail
@@ -293,8 +146,10 @@ async def lsp_symbols(
     section_range = Range(
         start=section_header_location.range.start,
         end=Position(
-            max(s.range.end.line for s in children
-                ) if children else section_header_location.range.end.line),
+            line=max(s.range.end.line for s in children)
+            if children else section_header_location.range.end.line,
+            character=0,
+        ),
     )
 
     symbols.append(
@@ -309,14 +164,14 @@ async def lsp_symbols(
   return symbols
 
 
-@server.feature(COMPLETION, trigger_characters=["{", ":"])
+@server.feature(COMPLETION, CompletionOptions(trigger_characters=["{", ":"]))
 async def lsp_completion(
     ls: LanguageServer,
     params: CompletionParams,
 ) -> Optional[List[CompletionItem]]:
   await asyncio.sleep(DEBOUNCE_DELAY)
   items: List[CompletionItem] = []
-  doc = ls.workspace.get_document(params.textDocument.uri)
+  doc = ls.workspace.get_document(params.text_document.uri)
 
   def getSectionReferenceCompletionTextEdit(
       doc: Document,
@@ -335,17 +190,18 @@ async def lsp_completion(
         start = match.start()
         end = match.end()
         return TextEdit(
-            Range(Position(pos.line, start), Position(pos.line, end)),
-            new_text,
+            range=Range(start=Position(line=pos.line, character=start),
+                        end=Position(line=pos.line, character=end)),
+            new_text=new_text,
         )
       index = max(match.start(), index + 1)
 
     return TextEdit(
         Range(
-            Position(pos.line, pos.character),
-            Position(pos.line, pos.character),
+            start=Position(line=pos.line, character=pos.character),
+            end=Position(line=pos.line, character=pos.character),
         ),
-        new_text,
+        new_text=new_text,
     )
 
   def getOptionReferenceTextEdit(
@@ -368,17 +224,18 @@ async def lsp_completion(
         start = match.start() + section_len
         end = match.end()
         return TextEdit(
-            Range(Position(pos.line, start), Position(pos.line, end)),
-            new_text,
+            range=Range(start=Position(line=pos.line, character=start),
+                        end=Position(line=pos.line, character=end)),
+            new_text=new_text,
         )
       index = max(match.start(), index + 1)
 
     return TextEdit(
-        Range(
-            Position(pos.line, pos.character),
-            Position(pos.line, pos.character),
+        range=Range(
+            start=Position(line=pos.line, character=pos.character),
+            end=Position(line=pos.line, character=pos.character),
         ),
-        new_text,
+        new_text=new_text,
     )
 
   def getDefaultTextEdit(
@@ -396,9 +253,11 @@ async def lsp_completion(
       line = doc.lines[pos.line]
     if not line.strip():
       return TextEdit(
-          Range(Position(pos.line, pos.character),
-                Position(pos.line, pos.character)),
-          new_text,
+          range=Range(
+              start=Position(line=pos.line, character=pos.character),
+              end=Position(line=pos.line, character=pos.character),
+          ),
+          new_text=new_text,
       )
     index = 0
     while True:
@@ -412,12 +271,13 @@ async def lsp_completion(
           end += 1
         # TODO: test
         return TextEdit(
-            Range(Position(pos.line, start), Position(pos.line, end)),
-            new_text,
+            range=Range(start=Position(line=pos.line, character=start),
+                        end=Position(line=pos.line, character=end)),
+            new_text=new_text,
         )
       index = max(match.start(), index + 1)
 
-  parsed = await buildout.open(ls, params.textDocument.uri)
+  parsed = await buildout.open(ls, params.text_document.uri)
   if parsed is None:
     return None
   symbol = await parsed.getSymbolAtPosition(params.position)
@@ -428,7 +288,7 @@ async def lsp_completion(
         documentation = '```ini\n{}\n```'.format(
             '\n'.join('{} = {}'.format(k, v.value)
                       for (k, v) in section_items.items()
-                      if v and not v.implicit_option), )
+                      if v and not v.default_value), )
         if section_items.get('recipe'):
           recipe = section_items.getRecipe()
           if recipe:
@@ -489,24 +349,31 @@ async def lsp_completion(
       #   opt|
       assert isinstance(parsed, buildout.BuildoutProfile)
 
-      # if section is buildout, completes with buildout known options.
+      # complete with existing options from this sections, to override
+      # options and for [buildout]
+      for option_name, option_default_value in symbol.current_section.items():
+        # skip some options that are not supposed to be defined, only referenced
+        if option_name in (
+            '_buildout_section_name_',
+            '_profile_base_location_',
+        ):
+          continue
+        items.append(
+            CompletionItem(label=option_name,
+                           text_edit=getDefaultTextEdit(
+                               doc,
+                               params.position,
+                               option_name + ' = ',
+                           ),
+                           kind=CompletionItemKind.Variable,
+                           documentation=MarkupContent(
+                               kind=MarkupKind.Markdown,
+                               value=f'`{option_default_value.value}`',
+                           )))
+
+      # if section is buildout, completes extends & parts which are usually
+      # multi lines already with an extra \n
       if symbol.current_section_name == 'buildout':
-        # buildout default options
-        for option_name, option_default_value in parsed['buildout'].items():
-          items.append(
-              CompletionItem(label=option_name,
-                             text_edit=getDefaultTextEdit(
-                                 doc,
-                                 params.position,
-                                 option_name + ' = ',
-                             ),
-                             kind=CompletionItemKind.Variable,
-                             documentation=MarkupContent(
-                                 kind=MarkupKind.Markdown,
-                                 value=f'`{option_default_value.value}`',
-                             )))
-        # extra buildout options that are usually defined as multi line
-        # (so we insert a \n)
         for option_name, option_documentation in (
             ('extends', 'Profiles extended by this buildout'),
             ('parts', 'Parts that will be installed'),
@@ -629,7 +496,7 @@ async def lsp_definition(
     params: TextDocumentPositionParams,
 ) -> List[Location]:
   await asyncio.sleep(DEBOUNCE_DELAY)
-  parsed = await buildout.open(ls, params.textDocument.uri)
+  parsed = await buildout.open(ls, params.text_document.uri)
   if parsed is None:
     return []
   symbol = await parsed.getSymbolAtPosition(params.position)
@@ -657,11 +524,12 @@ async def lsp_definition(
       elif symbol.current_section_name == 'buildout' and symbol.current_option_name == 'extends':
         extend = symbol.value
         if not buildout._isurl(extend):
-          uri = params.textDocument.uri
+          uri = params.text_document.uri
           base = uri[:uri.rfind('/')] + '/'
           locations.append(
               Location(uri=urllib.parse.urljoin(base, extend),
-                       range=Range(start=Position(0, 0), end=Position(1, 0))))
+                       range=Range(start=Position(line=0, character=0),
+                                   end=Position(line=1, character=0))))
   return locations
 
 
@@ -671,7 +539,7 @@ async def lsp_references(
     params: TextDocumentPositionParams,
 ) -> List[Location]:
   references: List[Location] = []
-  searched_document = await buildout.parse(server, params.textDocument.uri)
+  searched_document = await buildout.parse(server, params.text_document.uri)
   assert searched_document is not None
   searched_symbol = await searched_document.getSymbolAtPosition(params.position
                                                                 )
@@ -703,18 +571,20 @@ async def lsp_references(
           for option_text, option_range in profile.getOptionValues(
               'buildout', 'parts'):
             if searched_section == option_text:
-              references.append(Location(profile.uri, option_range))
+              references.append(Location(uri=profile.uri, range=option_range))
 
         async for symbol in profile.getAllOptionReferenceSymbols():
           if symbol.referenced_section_name == searched_section:
             if searched_option is None:
-              references.append(Location(profile.uri, symbol.section_range))
+              references.append(
+                  Location(uri=profile.uri, range=symbol.section_range))
             elif symbol.referenced_option_name == searched_option:
-              references.append(Location(profile.uri, symbol.option_range))
+              references.append(
+                  Location(uri=profile.uri, range=symbol.option_range))
 
         if searched_option is None:
           # find references in <= macros
-          for section, options in profile.items():
+          for options in profile.values():
             for option_key, option_value in options.items():
               if option_key == '<':
                 if option_value.value == searched_section:
@@ -730,7 +600,7 @@ async def lsp_hover(
     params: TextDocumentPositionParams,
 ) -> Optional[Hover]:
   await asyncio.sleep(DEBOUNCE_DELAY)
-  parsed = await buildout.open(ls, params.textDocument.uri)
+  parsed = await buildout.open(ls, params.text_document.uri)
   if parsed is None:
     return None
   symbol = await parsed.getSymbolAtPosition(params.position)
@@ -755,7 +625,7 @@ async def lsp_document_link(
 ) -> List[DocumentLink]:
   await asyncio.sleep(DEBOUNCE_DELAY)
   links: List[DocumentLink] = []
-  uri = params.textDocument.uri
+  uri = params.text_document.uri
   parsed_buildout = await buildout.parse(ls, uri)
   base = uri[:uri.rfind('/')] + '/'
 
